package fr.enssat.lancer;

import static org.assertj.core.api.Assertions.*;
import org.junit.jupiter.api.*;

/**
 * classe permettant de vérifier quelques fonctions de base du logiciel
 * NB: c'est du JUnit 5 avec AssertJ
 *
 * Pour exécuter ces tests :
 *  - dans Eclipse : cliquer droit et choisir "Run as..." puis JUnit Test
 *  - en ligne de commande : mvn test
 *
 * @author pierre
 *
 */
public class Tests
{
    @Test
    public void testVectorConstructor()
    {
        // test de la construction d'un vecteur par deux points

        // arrange
        Point p1 = new Point(4,-6,2);
        Point p2 = new Point(5,-3,7);

        // act
        Vecteur V = new Vecteur(p1, p2);

        // assert
        assertThat(V.x).isCloseTo( 1.0f, within(Constantes.EPSILON));
        assertThat(V.y).isCloseTo( 3.0f, within(Constantes.EPSILON));
        assertThat(V.z).isCloseTo( 5.0f, within(Constantes.EPSILON));
    }

    @Test
    public void testVectorSubtract1()
    {
        // test de la soustraction de deux Vecteur

        // arrange
        Vecteur v1 = new Vecteur(5,-3,2);
        Vecteur v2 = new Vecteur(4,-6,7);

        // act
        Vecteur sub = Vecteur.sub(v1, v2);

        // assert
        assertThat(sub.x).isCloseTo( 1.0f, within(Constantes.EPSILON));
        assertThat(sub.y).isCloseTo( 3.0f, within(Constantes.EPSILON));
        assertThat(sub.z).isCloseTo(-5.0f, within(Constantes.EPSILON));
    }

    @Test
    public void testVectorSubtract2()
    {
        // test de la soustraction de deux Vecteur

        // arrange
        Vecteur v1 = new Vecteur(5,-3,2);
        Vecteur v2 = new Vecteur(4,-6,7);

        // act
        Vecteur sub = v1.sub(v2);

        // assert
        assertThat(sub.x).isCloseTo( 1.0f, within(Constantes.EPSILON));
        assertThat(sub.y).isCloseTo( 3.0f, within(Constantes.EPSILON));
        assertThat(sub.z).isCloseTo(-5.0f, within(Constantes.EPSILON));
    }

    @Test
    public void testVectorDot1()
    {
        // test du produit scalaire

        // arrange
        Vecteur v1 = new Vecteur(-2,-3,-4);
        Vecteur v2 = new Vecteur(5,6,7);

        // act
        float ps = v1.dot(v2);

        // assert
        assertThat(ps).isCloseTo(-56.0f, within(Constantes.EPSILON));
    }

    @Test
    public void testVectorDot2()
    {
        // test du produit scalaire

        // arrange
        Vecteur v1 = new Vecteur(-2,-3,-4);
        Vecteur v2 = new Vecteur(5,6,7);

        // act
        float ps = v2.dot(v1);

        // assert
        assertThat(ps).isCloseTo(-56.0f, within(Constantes.EPSILON));
    }


    @Test
    public void testVectorNorme2()
    {
        // test de la normalisation

        // arrange
        Vecteur v1 = new Vecteur(2,-3,4);

        // act
        float n2 = v1.norme2();

        // assert
        assertThat(n2).isCloseTo(29f, within(Constantes.EPSILON));
    }


    @Test
    public void testVectorNorme()
    {
        // test de la normalisation

        // arrange
        Vecteur v1 = new Vecteur(2,-3,4);

        // act
        float n = v1.norme();

        // assert
        assertThat(n).isCloseTo(5.385164f, within(Constantes.EPSILON));
    }


    @Test
    public void testVectorNormalized()
    {
        // test de la normalisation

        // arrange
        Vecteur v1 = new Vecteur(2,-3,4);

        // act
        Vecteur V = v1.normalized();

        // assert
        assertThat(V.x).isCloseTo( 0.371391f, within(Constantes.EPSILON));
        assertThat(V.y).isCloseTo(-0.557086f, within(Constantes.EPSILON));
        assertThat(V.z).isCloseTo( 0.742781f, within(Constantes.EPSILON));
    }

    @Test
    public void testVectorNormalize()
    {
        // test de la normalisation

        // arrange
        Vecteur V = new Vecteur(2,3,-4);

        // act
        Vecteur.normalize(V);

        // assert
        assertThat(V.x).isCloseTo( 0.371391f, within(Constantes.EPSILON));
        assertThat(V.y).isCloseTo( 0.557086f, within(Constantes.EPSILON));
        assertThat(V.z).isCloseTo(-0.742781f, within(Constantes.EPSILON));
    }

    @Test
    public void testRayonConstructor()
    {
        // test du constructeur d'un rayon

        // arrange
        Point p1 = new Point(2,-3,4);
        Point p2 = new Point(5,6,-7);

        // act
        Rayon R = new Rayon(p1, p2);

        // assert
        assertThat(R.P.x).isCloseTo( 2.0f, within(Constantes.EPSILON));
        assertThat(R.P.y).isCloseTo(-3.0f, within(Constantes.EPSILON));
        assertThat(R.P.z).isCloseTo( 4.0f, within(Constantes.EPSILON));

        assertThat(R.V.x).isCloseTo( 0.206529f, within(Constantes.EPSILON));
        assertThat(R.V.y).isCloseTo( 0.619586f, within(Constantes.EPSILON));
        assertThat(R.V.z).isCloseTo(-0.757271f, within(Constantes.EPSILON));
    }

    @Test
    public void testChercherIntersection1()
    {
        // test de ChercherIntersection

        // arrange
        Sphere sphere = new Sphere(new Point(0,0,10), 2.0f);
        Rayon R1 = new Rayon(new Point(0,0,-10), new Point(0.95f,0,0));

        // act
        float d1 = sphere.getIntersectionDistance(R1);

        // assert
        assertThat(d1).isCloseTo(19.26051f, within(Constantes.EPSILON));
    }


    @Test
    public void testChercherIntersection2()
    {
        // test de ChercherIntersection

        // arrange
        Sphere sphere = new Sphere(new Point(0,0,10), 2.0f);
        Rayon R2 = new Rayon(new Point(2,0,-10), new Point(2.001f,0,0));

        // act
        float d2 = sphere.getIntersectionDistance(R2);

        // assert
        assertThat(d2).isGreaterThan(1e9f);
    }

}
